@@ -33,8 +33,10 @@ from typing import Any, Callable, Dict, Iterable, Iterator, List, Optional, Set,
 
 # 3rd party
 from domdf_python_tools.compat import importlib_metadata
+from domdf_python_tools.doctools import prettify_docstrings
+from domdf_python_tools.iterative import natmax, natmin
 from domdf_python_tools.paths import PathPlus
-from domdf_python_tools.stringlist import StringList
+from domdf_python_tools.stringlist import DelimitedList, StringList
 from domdf_python_tools.typing import PathLike
 from packaging.markers import default_environment
 from packaging.requirements import InvalidRequirement, Requirement
@@ -54,22 +56,11 @@ __all__ = [
 		"check_dependencies",
 		]
 
-
-def _check_equal_not_none(left: Optional[Any], right: Optional[Any]):
-	if not left or not right:
-		return True
-	else:
-		return left == right
-
-
-def _check_marker_equality(left: Optional[Any], right: Optional[Any]):
-	if left is not None and right is not None:
-		for left_mark, right_mark in zip(left._markers, right._markers):
-			if str(left_mark) != str(right_mark):
-				return False
-	return True
+operator_symbols = ("<=", '<', "!=", "==", ">=", '>', "~=", "===")
+_Requirement = Union[str, Requirement]
 
 
+@prettify_docstrings
 class ComparableRequirement(Requirement):
 	"""
 	Represents a :pep:`508` requirement.
@@ -78,6 +69,21 @@ class ComparableRequirement(Requirement):
 	A list of :class:`~.ComparableRequirement` objects can be sorted alphabetically.
 	"""
 
+	@staticmethod
+	def _check_equal_not_none(left: Optional[Any], right: Optional[Any]):
+		if not left or not right:
+			return True
+		else:
+			return left == right
+
+	@staticmethod
+	def _check_marker_equality(left: Optional[Any], right: Optional[Any]):
+		if left is not None and right is not None:
+			for left_mark, right_mark in zip(left._markers, right._markers):
+				if str(left_mark) != str(right_mark):
+					return False
+		return True
+
 	def __eq__(self, other) -> bool:
 
 		if isinstance(other, str):
@@ -90,47 +96,59 @@ class ComparableRequirement(Requirement):
 
 		elif isinstance(other, Requirement):
 			return all((
-					_check_equal_not_none(self.name, other.name),
-					_check_equal_not_none(self.url, other.url),
-					_check_equal_not_none(self.extras, other.extras),
-					_check_equal_not_none(self.specifier, other.specifier),
-					_check_marker_equality(self.marker, other.marker),
+					self._check_equal_not_none(self.name, other.name),
+					self._check_equal_not_none(self.url, other.url),
+					self._check_equal_not_none(self.extras, other.extras),
+					self._check_equal_not_none(self.specifier, other.specifier),
+					self._check_marker_equality(self.marker, other.marker),
 					))
 		else:  # pragma: no cover
 			return NotImplemented
 
-	def __gt__(self, other) -> bool:
-		if isinstance(other, Requirement):
-			return self.name > other.name
-		elif isinstance(other, str):
-			return self.name > other
-		else:  # pragma: no cover
-			return NotImplemented
-
-	def __ge__(self, other) -> bool:
-		if isinstance(other, Requirement):
-			return self.name >= other.name
-		elif isinstance(other, str):
-			return self.name >= other
-		else:  # pragma: no cover
-			return NotImplemented
-
-	def __le__(self, other) -> bool:
-		if isinstance(other, Requirement):
-			return self.name <= other.name
-		elif isinstance(other, str):
-			return self.name <= other
-		else:  # pragma: no cover
-			return NotImplemented
-
 	def __lt__(self, other) -> bool:
 		if isinstance(other, Requirement):
-			return self.name < other.name
+			if self.name != other.name:
+				return self.name < other.name
+			else:
+				if str(self.specifier or '') != str(other.specifier or ''):
+					return str(self.specifier or '') > str(other.specifier or '')
+				else:
+					return str(self.marker or '') > str(other.marker or '')
+
 		elif isinstance(other, str):
 			return self.name < other
 		else:  # pragma: no cover
 			return NotImplemented
 
+	def __le__(self, other) -> bool:
+		if not isinstance(other, (Requirement, str)):  # pragma: no cover
+			return NotImplemented
+		if self < other or self == other:
+			return True
+		return False
+
+	def __ge__(self, other) -> bool:
+		if not isinstance(other, (Requirement, str)):  # pragma: no cover
+			return NotImplemented
+		if self > other or self == other:
+			return True
+		return False
+
+	def __gt__(self, other) -> bool:
+		if isinstance(other, Requirement):
+			if self.name != other.name:
+				return self.name > other.name
+			else:
+				if str(self.specifier or '') != str(other.specifier or ''):
+					return str(self.specifier or '') < str(other.specifier or '')
+				else:
+					return str(self.marker or '') < str(other.marker or '')
+
+		elif isinstance(other, str):
+			return self.name > other
+		else:  # pragma: no cover
+			return NotImplemented
+
 	def __hash__(self) -> int:
 		return hash((
 				self.name or '',
@@ -141,7 +159,22 @@ class ComparableRequirement(Requirement):
 				))
 
 
-operator_symbols = ("<=", '<', "!=", "==", ">=", '>', "~=", "===")
+class _OperatorLookup(Dict[str, DelimitedList[Specifier]]):
+
+	def __setitem__(self, key: str, value):
+		if key not in operator_symbols:
+			raise KeyError(f"Invalid operator symbol {key!r}")
+
+		if isinstance(value, DelimitedList):
+			super().__setitem__(key, value)
+		else:
+			super().__setitem__(key, DelimitedList(value))
+
+	def __getitem__(self, item: str) -> DelimitedList[Specifier]:
+		if item not in self and item in operator_symbols:
+			self[item] = DelimitedList()
+
+		return super().__getitem__(item)
 
 
 def resolve_specifiers(specifiers: Iterable[Specifier]) -> SpecifierSet:
@@ -151,44 +184,82 @@ def resolve_specifiers(specifiers: Iterable[Specifier]) -> SpecifierSet:
 	:param specifiers:
 	"""
 
-	final_specifier_set = SpecifierSet()
-
-	operator_lookup: Dict[str, List[Specifier]] = {s: [] for s in operator_symbols}
+	operator_lookup = _OperatorLookup()
+	spec: Specifier
 
 	for spec in specifiers:
-		if spec.operator in operator_lookup:
+		if spec.operator in operator_symbols:
 			operator_lookup[spec.operator].append(spec)
 
 	if operator_lookup["<="]:
-		final_specifier_set &= SpecifierSet(f"<={min(spec.version for spec in operator_lookup['<='])}")
+		operator_lookup["<="] = [Specifier(f"<={natmin(spec.version for spec in operator_lookup['<='])}")]
 
 	if operator_lookup['<']:
-		final_specifier_set &= SpecifierSet(f"<{min(spec.version for spec in operator_lookup['<'])}")
-
-	for spec in operator_lookup["!="]:
-		final_specifier_set &= SpecifierSet(f"!={spec.version}")
-
-	for spec in operator_lookup["=="]:
-		final_specifier_set &= SpecifierSet(f"=={spec.version}")
+		operator_lookup['<'] = [Specifier(f"<{natmin(spec.version for spec in operator_lookup['<'])}")]
 
 	if operator_lookup[">="]:
-		final_specifier_set &= SpecifierSet(f">={max(spec.version for spec in operator_lookup['>='])}")
+		operator_lookup[">="] = [Specifier(f">={natmax(spec.version for spec in operator_lookup['>='])}")]
 
 	if operator_lookup['>']:
-		final_specifier_set &= SpecifierSet(f">{max(spec.version for spec in operator_lookup['>'])}")
+		operator_lookup['>'] = [Specifier(f">{natmax(spec.version for spec in operator_lookup['>'])}")]
+
+	# merge e.g. >1.2.3 and >=1.2.2 into >1.2.3
+	if operator_lookup[">="] and operator_lookup['>']:
+		gt_version = operator_lookup['>'][0].version
+		ge_version = operator_lookup[">="][0].version
+
+		if gt_version > ge_version:
+			del operator_lookup[">="]
+
+	# merge e.g. >=1.2.2 and ==1.2.3 into ==1.2.3
+	if operator_lookup[">="] and operator_lookup["=="]:
+		ge_version = operator_lookup[">="][0].version
+
+		if any([eq_version.version >= ge_version for eq_version in operator_lookup["=="]]):
+			del operator_lookup[">="]
+
+	# merge e.g. <=1.2.3 and <1.2.2 into <1.2.2
+	if operator_lookup["<="] and operator_lookup['<']:
+		lt_version = operator_lookup['<'][0].version
+		le_version = operator_lookup["<="][0].version
+
+		if lt_version < le_version:
+			del operator_lookup["<="]
+
+	# merge e.g. <=1.2.3 and ==1.2.2 into ==1.2.2
+	if operator_lookup["<="] and operator_lookup["=="]:
+		le_version = operator_lookup["<="][0].version
+
+		if any([eq_version.version <= le_version for eq_version in operator_lookup["=="]]):
+			del operator_lookup["<="]
+
+	specifier_set = SpecifierSet()
+
+	if operator_lookup["<="]:
+		specifier_set &= SpecifierSet(f"{operator_lookup['<=']:,}")
+
+	if operator_lookup['<']:
+		specifier_set &= SpecifierSet(f"{operator_lookup['<']:,}")
+
+	for spec in operator_lookup["!="]:
+		specifier_set &= SpecifierSet(f"!={spec.version}")
+
+	for spec in operator_lookup["=="]:
+		specifier_set &= SpecifierSet(f"=={spec.version}")
+
+	if operator_lookup[">="]:
+		specifier_set &= SpecifierSet(f"{operator_lookup['>=']:,}")
+
+	if operator_lookup['>']:
+		specifier_set &= SpecifierSet(f"{operator_lookup['>']:,}")
 
 	for spec in operator_lookup["~="]:
-		final_specifier_set &= SpecifierSet(f"~={spec.version}")
+		specifier_set &= SpecifierSet(f"~={spec.version}")
 
 	for spec in operator_lookup["==="]:
-		final_specifier_set &= SpecifierSet(f"==={spec.version}")
+		specifier_set &= SpecifierSet(f"==={spec.version}")
 
-	# TODO: merge e.g. >1.2.3 and >=1.2.2 (into >1.2.3)
-
-	return final_specifier_set
-
-
-_Requirement = Union[str, Requirement]
+	return specifier_set
 
 
 def combine_requirements(
@@ -201,10 +272,11 @@ def combine_requirements(
 
 	:param requirement: A single requirement, or an iterable of requirements.
 	:param requirements: Additional requirements.
+	:param normalize_func: Function to use to normalize the names of requirements.
 
-	.. versionchanged:: 0.2.1 Added the ``normalize_func`` keyword-only argument.
+	.. versionchanged:: 0.2.1  Added the ``normalize_func`` keyword-only argument.
 
-	.. TODO:: Markers
+	.. versionchanged:: 0.3.1  Requirements are no longer combined if their markers differ.
 	"""
 
 	if isinstance(requirement, Iterable):
@@ -222,13 +294,13 @@ def combine_requirements(
 
 		if req.name in merged_requirements:
 			other_req = merged_requirements[merged_requirements.index(req.name)]  # type: ignore
-			other_req.specifier &= req.specifier
-			other_req.extras &= req.extras
-			other_req.specifier = resolve_specifiers(other_req.specifier)
-			if req.marker and other_req.marker:
-				raise NotImplementedError
-			elif req.marker and not other_req.marker:
-				other_req.marker = req.marker
+
+			if req.marker != other_req.marker:
+				merged_requirements.append(req)
+			else:
+				other_req.specifier &= req.specifier
+				other_req.extras &= req.extras
+				other_req.specifier = resolve_specifiers(other_req.specifier)
 		else:
 			merged_requirements.append(req)
 
@@ -244,6 +316,8 @@ _read_requirements_ret = Union[_read_requirements_ret_invalid, _read_requirement
 def read_requirements(
 		req_file: PathLike,
 		include_invalid: Literal[True],
+		*,
+		normalize_func: Callable[[str], str] = ...
 		) -> _read_requirements_ret_invalid:
 	...  # pragma: no cover
 
@@ -252,6 +326,8 @@ def read_requirements(
 def read_requirements(
 		req_file: PathLike,
 		include_invalid: Literal[False] = ...,
+		*,
+		normalize_func: Callable[[str], str] = ...
 		) -> _read_requirements_ret_valid:
 	...  # pragma: no cover
 
@@ -281,21 +357,22 @@ def read_requirements(
 	requirements: Set[ComparableRequirement] = set()
 
 	for line in PathPlus(req_file).read_lines():
-		if line.startswith("#"):
+		if line.startswith('#'):
 			comments.append(line)
 		elif line:
 			try:
 				req = ComparableRequirement(line)
 				req.name = normalize_func(req.name)
-				if req.name not in [normalize_func(r.name) for r in requirements]:
-					requirements.add(req)
+				requirements.add(req)
 			except InvalidRequirement:
-				warnings.warn(f"Ignored invalid requirement {line!r}")
 				invalid_lines.append(line)
 
 	if include_invalid:
 		return requirements, comments, invalid_lines
 	else:
+		for line in invalid_lines:
+			warnings.warn(f"Ignored invalid requirement {line!r}")
+
 		return requirements, comments
 
 
@@ -313,8 +390,12 @@ class RequirementsManager(ABC):
 	:param repo_path: Path to the repository root.
 	"""
 
-	#: The static target requirements
-	target_requirements: Set[Requirement]
+	target_requirements: Set[ComparableRequirement]
+	"""
+	The static target requirements
+
+	.. versionchanged:: 0.4.0  Previously this was a set of :class:`packaging.requirements.Requirement`.
+	"""
 
 	#: The path of the requirements file, relative to the repository root.
 	filename: PathLike
@@ -322,6 +403,7 @@ class RequirementsManager(ABC):
 	def __init__(self, repo_path: PathLike):
 		self.repo_path = PathPlus(repo_path)
 		self.req_file = self.prep_req_file()
+		self.target_requirements = set(self.target_requirements)
 
 	def prep_req_file(self):
 		"""
@@ -389,8 +471,18 @@ class RequirementsManager(ABC):
 		"""
 
 		lib_requirements, _ = read_requirements(self.repo_path / "requirements.txt")
-		lib_requirements_names = [self.normalize(r.name) for r in lib_requirements]
-		self.target_requirements = {r for r in self.target_requirements if r.name not in lib_requirements_names}
+		lib_requirements_names_extras = {normalize(r.name): r.extras for r in lib_requirements}
+
+		non_library_requirements = set()
+
+		for req in self.target_requirements:
+			if req.name in lib_requirements_names_extras:
+				if req.extras != lib_requirements_names_extras[req.name]:
+					non_library_requirements.add(req)
+			else:
+				non_library_requirements.add(req)
+
+		self.target_requirements = non_library_requirements
 
 	def write_requirements(self, comments: List[str]) -> None:
 		"""
@@ -403,7 +495,7 @@ class RequirementsManager(ABC):
 
 		buf = StringList(comments)
 
-		for req in sorted(self.target_requirements, key=lambda r: r.name.casefold()):
+		for req in sorted(self.target_requirements):
 			buf.append(str(req))
 
 		self.req_file.write_lines(buf)
